import os
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface


class SQLGenerator(GeneratorInterface):
    """
    SQLGenerator is a class that implements the GeneratorInterface and produces the code or set of SQL statements 
    used to define and modify the structure of the tables in a database..

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
        sql_dialect (str, optional): The SQL dialect. Values allowed: None, "postgres", or "mysql".
    """

    TYPES = {
        "int": "int",
        "str": "varchar(100)",
        "float": "float8",
        "bool": "bool",
        "time": "time",
        "date": "date",
        "datetime": "timestamp",
        "timedelta": "interval",
    }

    def __init__(self, model: DomainModel, output_dir: str = None, sql_dialect: str = None):
        super().__init__(model, output_dir)
        self.sql_dialect = sql_dialect

    def generate(self):
        """
        Generates SQL code based on the provided B-UML model and saves it to the specified output directory.
        If the output directory was not specified, the code generated will be stored in the <current directory>/output
        folder.

        Returns:
            None, but store the generated code as a file named sql_alchemy.py 
        """
        file_path = self.build_generation_path(file_name="tables.sql")
        templates_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), "templates")
        env = Environment(loader=FileSystemLoader(templates_path), trim_blocks=True, lstrip_blocks=True)
        template = env.get_template('sql_template.sql.j2')
        with open(file_path, mode="w") as f:
            generated_code = template.render(model=self.model, types=self.TYPES, sql_dialect=self.sql_dialect)
            f.write(generated_code)
            print("Code generated in the location: " + file_path)



import pytest

from besser.BUML.metamodel.structural import NamedElement, DomainModel, Type, Class, \
    Property, PrimitiveDataType, Multiplicity, Association, BinaryAssociation, Generalization, \
    GeneralizationSet, AssociationClass, Enumeration, EnumerationLiteral


def test_named_element():
    named_element: NamedElement = NamedElement(name="element1")
    assert named_element.name == "element1"


def test_model_initialization():
    class1: Type = Type(name="element1")
    class2: Type = Type(name="element2")
    model: DomainModel = DomainModel(name="mymodel", types={class1, class2}, associations = None, packages = None, constraints = None)
    assert len(model.types) == 2
    model_empty: DomainModel = DomainModel(name="mymodel", types = None, associations = None, packages = None, constraints = None)
    assert len(model_empty.types) == 0


# Testing the WFR for duplicate names in a model
def test_model_duplicated_names():
    with pytest.raises(ValueError) as excinfo:
        class1: Type = Type(name="name1")
        class2: Type = Type(name="name1")
        model: DomainModel = DomainModel(name="mymodel", types={class1, class2}, associations = None, packages = None, constraints = None)
    assert "same name" in str(excinfo.value)


# Testing attributes initialization
def test_attribute_initialization():
    class1: Type = Type(name="name1")
    attribute1: Property = Property(name="attribute1", owner = class1, type=PrimitiveDataType("int"),
                                    multiplicity=Multiplicity(0, 1))
    # assert attributes has proper type and multiplicity
    assert attribute1.type.name == "int"
    assert attribute1.multiplicity.min == 0
    assert attribute1.multiplicity.max == 1


# Testing attribute multiplicity and data type violations
def test_attribute_type_and_multiplicity_violation():
    with pytest.raises(ValueError) as excinfo:
        attribute1: Property = Property(name="attribute1", type=PrimitiveDataType("int"),
                                        multiplicity=Multiplicity(0, -1))
    assert "Invalid max multiplicity" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        attribute1: Property = Property(name="attribute1", type=PrimitiveDataType("int"),
                                        multiplicity=Multiplicity(-1, 1))
    assert "Invalid min multiplicity" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        attribute1: Property = Property(name="attribute1", type=PrimitiveDataType("int"),
                                        multiplicity=Multiplicity(2, 1))
    assert "Invalid max multiplicity" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        attribute1: Property = Property(name="attribute1", type=PrimitiveDataType("invented_type"),
                                        multiplicity=Multiplicity(2, 1))
    assert "Invalid primitive data type" in str(excinfo.value)


# Testing class initialization
def test_class_initialization():
    class1: Class = Class(name="classA", attributes=set())
    class2: Class
    attribute1: Property = Property(name="attribute1", owner=None, type=PrimitiveDataType("int"), multiplicity=Multiplicity(0, 1))
    attribute2: Property = Property(name="attribute2", owner=None, type=PrimitiveDataType("int"), multiplicity=Multiplicity(0, 1))
    reference1: Property = Property(name="reference1", owner=None, type=class1, multiplicity=Multiplicity(0, 1))
    class2: Class = Class(name="classB", attributes={attribute1, attribute2})
    assert len(class2.attributes) == 2
    print(class2)


# Testing no duplicated names in class attributes
def test_duplicated_name_class():
    with pytest.raises(ValueError) as excinfo:
        class1 : Class
        attribute1: Property = Property(name="attribute1", owner=None, type=PrimitiveDataType("int"), multiplicity=Multiplicity(0, 1))
        attribute2: Property = Property(name="attribute1", owner=None, type=PrimitiveDataType("int"), multiplicity=Multiplicity(0, 1))
        class1 = Class(name="name1", attributes={attribute1, attribute2})
    assert "A class cannot have two attributes with the same name" in str(excinfo.value)

# Testing for no more than one id attribute in class
def test_more_than_one_id_class():
    with pytest.raises(ValueError) as excinfo:
        class1 : Class
        attribute1: Property = Property(name="attribute1", type=PrimitiveDataType("int"), is_id=True)
        attribute2: Property = Property(name="attribute2", type=PrimitiveDataType("int"), is_id=True)
        class1 = Class(name="name1", attributes={attribute1, attribute2})
    assert "A class cannot have two id attributes" in str(excinfo.value)

def test_association_initialization():
    class1: Class = Class(name="name1", attributes=set())
    class2: Class = Class(name="name2", attributes=set())
    aend1: Property = Property(name="end1", owner=None, type=class1, multiplicity=Multiplicity(0, 1))
    aend2: Property = Property(name="end2", owner=None, type=class2, multiplicity=Multiplicity(0, 1))
    association: BinaryAssociation = BinaryAssociation(name="association1", ends={aend1, aend2})
    assert len(association.ends) == 2
    assert aend1 in association.ends
    assert aend1.owner == association
    assert class1.associations == {association}
    assert class1.association_ends() == {aend2}

# Testing the creation of a binary association cannot have more than two ends
def test_binary_association():
    with pytest.raises(ValueError) as excinfo:
        class1: Type = Type(name="name1")
        aend: Property = Property(name="end1", owner=None, type=class1, multiplicity=Multiplicity(0, 1))
        association: BinaryAssociation = BinaryAssociation(name="association1", ends={aend})
    assert "A binary association must have exactly two ends" in str(excinfo.value)


# Testing the creation of an association class with an attribute
def test_association_class():
    class1: Class = Class(name="name1", attributes=None)
    class2: Class = Class(name="name2", attributes=None)
    aend1: Property = Property(name="end1", owner=None, type=class1, multiplicity=Multiplicity(0, 1))
    aend2: Property = Property(name="end2", owner=None, type=class2, multiplicity=Multiplicity(0, 1))
    association: BinaryAssociation = BinaryAssociation(name="association1", ends={aend1, aend2})
    attribute1: Property = Property(name="attribute1", owner=None, type=PrimitiveDataType("int"), multiplicity=Multiplicity(0, 1))
    association_class: AssociationClass = AssociationClass(name="association_class1", attributes={attribute1}, association=association)
    assert len(association_class.attributes) == 1
    assert attribute1 in association_class.attributes
    assert association_class.association.name == "association1"


def test_generalization_initialization():
    attribute1: Property = Property(name="attribute1", owner = None, type=PrimitiveDataType("int"),
                                    multiplicity=Multiplicity(0, 1))
    class1: Class = Class(name="name1", attributes={attribute1})
    class2: Class = Class(name="name2", attributes=None)
    generalization: Generalization = Generalization(general=class1, specific=class2)
    assert generalization.general == class1
    assert generalization.specific == class2
    assert class2.generalizations == {generalization}
    assert class1.specializations() == {class2}
    assert class2.parents() == {class1}
    assert class2.specializations() == set()
    assert class2.all_attributes() == {attribute1}


def test_no_generalization_loop():
    with pytest.raises(ValueError) as excinfo:
        class1: Class = Class(name="name1", attributes=None)
        generalization: Generalization = Generalization(general=class1, specific=class1)
    assert "A class cannot be a generalization of itself" in str(excinfo.value)

def test_generalization_set_initialization():
    class1: Class = Class(name="name1", attributes=None)
    class2: Class = Class(name="name2", attributes=None)
    class3: Class = Class(name="name3", attributes=None)
    generalization1: Generalization = Generalization(general=class1, specific=class2)
    generalization2: Generalization = Generalization(general=class1, specific=class3)
    generalization_set: GeneralizationSet = GeneralizationSet(name="Generalization Set", generalizations={
        generalization1,generalization2}, is_disjoint=True, is_complete=True)
    assert generalization_set.is_disjoint == True
    assert generalization_set.is_complete == True
    assert class1.generalizations == {generalization1, generalization2}
    assert class1.specializations() == {class3, class2}
    assert class2.parents() == {class1}
    assert class3.parents() == {class1}
    assert class2.specializations() == set()
    
# Testing enumeration and literals initialization
def test_enumeration_initialization():
    literal1: EnumerationLiteral = EnumerationLiteral(name="literal_1", owner=None)
    literal2: EnumerationLiteral = EnumerationLiteral(name="literal_2", owner=None)
    literal3: EnumerationLiteral = EnumerationLiteral(name="literal_3", owner=None)
    enum: Enumeration = Enumeration(name="Enumeration", literals={literal1,literal2,literal3})
    assert len(enum.literals) == 3
    assert literal1.owner == enum
    
# Testing no duplicated literal names in an enumeration
def test_duplicated_name_literal():
    with pytest.raises(ValueError) as excinfo:
        literal1: EnumerationLiteral = EnumerationLiteral(name="duplicated_name", owner=None)
        literal2: EnumerationLiteral = EnumerationLiteral(name="duplicated_name", owner=None)
        enum1: Enumeration = Enumeration(name="Enumeration", literals={literal1,literal2})
    assert "An enumeration cannot have two literals with the same name" in str(excinfo.value)



# Generated from ./PlantUML.g4 by ANTLR 4.13.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,40,200,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,2,11,7,11,2,12,7,12,2,13,7,13,
        2,14,7,14,2,15,7,15,2,16,7,16,2,17,7,17,2,18,7,18,2,19,7,19,1,0,
        1,0,1,0,5,0,44,8,0,10,0,12,0,47,9,0,1,0,1,0,1,1,1,1,1,1,3,1,54,8,
        1,1,2,1,2,1,2,1,2,1,2,1,3,1,3,3,3,63,8,3,1,3,1,3,3,3,67,8,3,1,3,
        1,3,1,3,1,3,5,3,73,8,3,10,3,12,3,76,9,3,1,3,1,3,1,3,1,4,1,4,3,4,
        83,8,4,1,5,1,5,3,5,87,8,5,1,6,1,6,3,6,91,8,6,1,6,1,6,1,6,1,6,3,6,
        97,8,6,1,6,3,6,100,8,6,1,6,1,6,1,6,3,6,105,8,6,1,6,1,6,1,7,1,7,1,
        8,3,8,112,8,8,1,8,1,8,3,8,116,8,8,1,9,3,9,119,8,9,1,9,3,9,122,8,
        9,3,9,124,8,9,1,9,1,9,3,9,128,8,9,1,9,3,9,131,8,9,3,9,133,8,9,1,
        10,3,10,136,8,10,1,10,3,10,139,8,10,3,10,141,8,10,1,10,1,10,3,10,
        145,8,10,1,10,3,10,148,8,10,3,10,150,8,10,1,11,1,11,1,11,3,11,155,
        8,11,1,11,1,11,1,11,1,12,1,12,1,12,1,13,1,13,1,13,1,13,3,13,167,
        8,13,1,13,1,13,1,14,1,14,1,15,3,15,174,8,15,1,15,1,15,1,15,1,15,
        1,15,1,16,3,16,182,8,16,1,16,3,16,185,8,16,1,16,3,16,188,8,16,1,
        16,1,16,1,16,1,16,1,17,1,17,1,18,1,18,1,19,1,19,1,19,0,0,20,0,2,
        4,6,8,10,12,14,16,18,20,22,24,26,28,30,32,34,36,38,0,4,1,0,36,37,
        1,0,19,22,1,0,23,30,1,0,31,32,214,0,40,1,0,0,0,2,53,1,0,0,0,4,55,
        1,0,0,0,6,62,1,0,0,0,8,80,1,0,0,0,10,86,1,0,0,0,12,88,1,0,0,0,14,
        108,1,0,0,0,16,111,1,0,0,0,18,123,1,0,0,0,20,140,1,0,0,0,22,151,
        1,0,0,0,24,159,1,0,0,0,26,162,1,0,0,0,28,170,1,0,0,0,30,173,1,0,
        0,0,32,181,1,0,0,0,34,193,1,0,0,0,36,195,1,0,0,0,38,197,1,0,0,0,
        40,41,5,33,0,0,41,45,5,40,0,0,42,44,3,2,1,0,43,42,1,0,0,0,44,47,
        1,0,0,0,45,43,1,0,0,0,45,46,1,0,0,0,46,48,1,0,0,0,47,45,1,0,0,0,
        48,49,5,34,0,0,49,1,1,0,0,0,50,54,3,4,2,0,51,54,3,6,3,0,52,54,3,
        10,5,0,53,50,1,0,0,0,53,51,1,0,0,0,53,52,1,0,0,0,54,3,1,0,0,0,55,
        56,5,1,0,0,56,57,5,2,0,0,57,58,5,36,0,0,58,59,5,40,0,0,59,5,1,0,
        0,0,60,63,3,8,4,0,61,63,5,3,0,0,62,60,1,0,0,0,62,61,1,0,0,0,63,64,
        1,0,0,0,64,66,5,35,0,0,65,67,3,24,12,0,66,65,1,0,0,0,66,67,1,0,0,
        0,67,68,1,0,0,0,68,69,5,4,0,0,69,74,5,40,0,0,70,73,3,30,15,0,71,
        73,3,32,16,0,72,70,1,0,0,0,72,71,1,0,0,0,73,76,1,0,0,0,74,72,1,0,
        0,0,74,75,1,0,0,0,75,77,1,0,0,0,76,74,1,0,0,0,77,78,5,5,0,0,78,79,
        5,40,0,0,79,7,1,0,0,0,80,82,5,6,0,0,81,83,5,3,0,0,82,81,1,0,0,0,
        82,83,1,0,0,0,83,9,1,0,0,0,84,87,3,12,6,0,85,87,3,22,11,0,86,84,
        1,0,0,0,86,85,1,0,0,0,87,11,1,0,0,0,88,90,5,35,0,0,89,91,3,26,13,
        0,90,89,1,0,0,0,90,91,1,0,0,0,91,96,1,0,0,0,92,97,3,14,7,0,93,97,
        3,16,8,0,94,97,3,18,9,0,95,97,3,20,10,0,96,92,1,0,0,0,96,93,1,0,
        0,0,96,94,1,0,0,0,96,95,1,0,0,0,97,99,1,0,0,0,98,100,3,26,13,0,99,
        98,1,0,0,0,99,100,1,0,0,0,100,101,1,0,0,0,101,104,5,35,0,0,102,103,
        5,7,0,0,103,105,5,35,0,0,104,102,1,0,0,0,104,105,1,0,0,0,105,106,
        1,0,0,0,106,107,5,40,0,0,107,13,1,0,0,0,108,109,5,8,0,0,109,15,1,
        0,0,0,110,112,5,9,0,0,111,110,1,0,0,0,111,112,1,0,0,0,112,113,1,
        0,0,0,113,115,5,8,0,0,114,116,5,10,0,0,115,114,1,0,0,0,115,116,1,
        0,0,0,116,17,1,0,0,0,117,119,5,11,0,0,118,117,1,0,0,0,118,119,1,
        0,0,0,119,124,1,0,0,0,120,122,5,9,0,0,121,120,1,0,0,0,121,122,1,
        0,0,0,122,124,1,0,0,0,123,118,1,0,0,0,123,121,1,0,0,0,124,125,1,
        0,0,0,125,132,5,8,0,0,126,128,5,10,0,0,127,126,1,0,0,0,127,128,1,
        0,0,0,128,133,1,0,0,0,129,131,5,11,0,0,130,129,1,0,0,0,130,131,1,
        0,0,0,131,133,1,0,0,0,132,127,1,0,0,0,132,130,1,0,0,0,133,19,1,0,
        0,0,134,136,5,37,0,0,135,134,1,0,0,0,135,136,1,0,0,0,136,141,1,0,
        0,0,137,139,5,9,0,0,138,137,1,0,0,0,138,139,1,0,0,0,139,141,1,0,
        0,0,140,135,1,0,0,0,140,138,1,0,0,0,141,142,1,0,0,0,142,149,5,8,
        0,0,143,145,5,10,0,0,144,143,1,0,0,0,144,145,1,0,0,0,145,150,1,0,
        0,0,146,148,5,37,0,0,147,146,1,0,0,0,147,148,1,0,0,0,148,150,1,0,
        0,0,149,144,1,0,0,0,149,147,1,0,0,0,150,21,1,0,0,0,151,154,5,35,
        0,0,152,155,5,12,0,0,153,155,5,13,0,0,154,152,1,0,0,0,154,153,1,
        0,0,0,155,156,1,0,0,0,156,157,5,35,0,0,157,158,5,40,0,0,158,23,1,
        0,0,0,159,160,5,14,0,0,160,161,5,35,0,0,161,25,1,0,0,0,162,163,5,
        15,0,0,163,166,3,28,14,0,164,165,5,16,0,0,165,167,3,28,14,0,166,
        164,1,0,0,0,166,167,1,0,0,0,167,168,1,0,0,0,168,169,5,15,0,0,169,
        27,1,0,0,0,170,171,7,0,0,0,171,29,1,0,0,0,172,174,3,34,17,0,173,
        172,1,0,0,0,173,174,1,0,0,0,174,175,1,0,0,0,175,176,5,35,0,0,176,
        177,5,7,0,0,177,178,3,36,18,0,178,179,5,40,0,0,179,31,1,0,0,0,180,
        182,3,34,17,0,181,180,1,0,0,0,181,182,1,0,0,0,182,184,1,0,0,0,183,
        185,3,38,19,0,184,183,1,0,0,0,184,185,1,0,0,0,185,187,1,0,0,0,186,
        188,5,17,0,0,187,186,1,0,0,0,187,188,1,0,0,0,188,189,1,0,0,0,189,
        190,5,35,0,0,190,191,5,18,0,0,191,192,5,40,0,0,192,33,1,0,0,0,193,
        194,7,1,0,0,194,35,1,0,0,0,195,196,7,2,0,0,196,37,1,0,0,0,197,198,
        7,3,0,0,198,39,1,0,0,0,32,45,53,62,66,72,74,82,86,90,96,99,104,111,
        115,118,121,123,127,130,132,135,138,140,144,147,149,154,166,173,
        181,184,187
    ]

class PlantUMLParser ( Parser ):

    grammarFileName = "PlantUML.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'skinparam'", "'groupInheritance'", "'class'", 
                     "'{'", "'}'", "'abstract'", "':'", "'--'", "'<'", "'>'", 
                     "'o'", "'<|--'", "'--|>'", "'extends'", "'\"'", "'..'", 
                     "'void'", "'()'", "'#'", "'-'", "'~'", "'+'", "'int'", 
                     "'float'", "'str'", "'bool'", "'time'", "'date'", "'datetime'", 
                     "'timedelta'", "'{static}'", "'{abstract}'", "'@startuml'", 
                     "'@enduml'", "<INVALID>", "<INVALID>", "'*'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "Start", "End", "ID", "INT", "ASTK", 
                      "DOUBLE_QUOTE", "WS", "NL" ]

    RULE_domainModel = 0
    RULE_element = 1
    RULE_skinParam = 2
    RULE_class = 3
    RULE_abstract = 4
    RULE_relationship = 5
    RULE_association = 6
    RULE_bidirectional = 7
    RULE_unidirectional = 8
    RULE_aggregation = 9
    RULE_composition = 10
    RULE_inheritance = 11
    RULE_extends = 12
    RULE_cardinality = 13
    RULE_cardinalityVal = 14
    RULE_attribute = 15
    RULE_method = 16
    RULE_visibility = 17
    RULE_primitiveData = 18
    RULE_modifier = 19

    ruleNames =  [ "domainModel", "element", "skinParam", "class", "abstract", 
                   "relationship", "association", "bidirectional", "unidirectional", 
                   "aggregation", "composition", "inheritance", "extends", 
                   "cardinality", "cardinalityVal", "attribute", "method", 
                   "visibility", "primitiveData", "modifier" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    T__10=11
    T__11=12
    T__12=13
    T__13=14
    T__14=15
    T__15=16
    T__16=17
    T__17=18
    T__18=19
    T__19=20
    T__20=21
    T__21=22
    T__22=23
    T__23=24
    T__24=25
    T__25=26
    T__26=27
    T__27=28
    T__28=29
    T__29=30
    T__30=31
    T__31=32
    Start=33
    End=34
    ID=35
    INT=36
    ASTK=37
    DOUBLE_QUOTE=38
    WS=39
    NL=40

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class DomainModelContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Start(self):
            return self.getToken(PlantUMLParser.Start, 0)

        def NL(self):
            return self.getToken(PlantUMLParser.NL, 0)

        def End(self):
            return self.getToken(PlantUMLParser.End, 0)

        def element(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PlantUMLParser.ElementContext)
            else:
                return self.getTypedRuleContext(PlantUMLParser.ElementContext,i)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_domainModel

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDomainModel" ):
                listener.enterDomainModel(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDomainModel" ):
                listener.exitDomainModel(self)




    def domainModel(self):

        localctx = PlantUMLParser.DomainModelContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_domainModel)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 40
            self.match(PlantUMLParser.Start)
            self.state = 41
            self.match(PlantUMLParser.NL)
            self.state = 45
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 34359738442) != 0):
                self.state = 42
                self.element()
                self.state = 47
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 48
            self.match(PlantUMLParser.End)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ElementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def skinParam(self):
            return self.getTypedRuleContext(PlantUMLParser.SkinParamContext,0)


        def class_(self):
            return self.getTypedRuleContext(PlantUMLParser.ClassContext,0)


        def relationship(self):
            return self.getTypedRuleContext(PlantUMLParser.RelationshipContext,0)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_element

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterElement" ):
                listener.enterElement(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitElement" ):
                listener.exitElement(self)




    def element(self):

        localctx = PlantUMLParser.ElementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_element)
        try:
            self.state = 53
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [1]:
                self.enterOuterAlt(localctx, 1)
                self.state = 50
                self.skinParam()
                pass
            elif token in [3, 6]:
                self.enterOuterAlt(localctx, 2)
                self.state = 51
                self.class_()
                pass
            elif token in [35]:
                self.enterOuterAlt(localctx, 3)
                self.state = 52
                self.relationship()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class SkinParamContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT(self):
            return self.getToken(PlantUMLParser.INT, 0)

        def NL(self):
            return self.getToken(PlantUMLParser.NL, 0)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_skinParam

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSkinParam" ):
                listener.enterSkinParam(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSkinParam" ):
                listener.exitSkinParam(self)




    def skinParam(self):

        localctx = PlantUMLParser.SkinParamContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_skinParam)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 55
            self.match(PlantUMLParser.T__0)
            self.state = 56
            self.match(PlantUMLParser.T__1)
            self.state = 57
            self.match(PlantUMLParser.INT)
            self.state = 58
            self.match(PlantUMLParser.NL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ClassContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def NL(self, i:int=None):
            if i is None:
                return self.getTokens(PlantUMLParser.NL)
            else:
                return self.getToken(PlantUMLParser.NL, i)

        def abstract(self):
            return self.getTypedRuleContext(PlantUMLParser.AbstractContext,0)


        def extends(self):
            return self.getTypedRuleContext(PlantUMLParser.ExtendsContext,0)


        def attribute(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PlantUMLParser.AttributeContext)
            else:
                return self.getTypedRuleContext(PlantUMLParser.AttributeContext,i)


        def method(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PlantUMLParser.MethodContext)
            else:
                return self.getTypedRuleContext(PlantUMLParser.MethodContext,i)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_class

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterClass" ):
                listener.enterClass(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitClass" ):
                listener.exitClass(self)




    def class_(self):

        localctx = PlantUMLParser.ClassContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_class)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 62
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [6]:
                self.state = 60
                self.abstract()
                pass
            elif token in [3]:
                self.state = 61
                self.match(PlantUMLParser.T__2)
                pass
            else:
                raise NoViableAltException(self)

            self.state = 64
            self.match(PlantUMLParser.ID)
            self.state = 66
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 65
                self.extends()


            self.state = 68
            self.match(PlantUMLParser.T__3)
            self.state = 69
            self.match(PlantUMLParser.NL)
            self.state = 74
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 40810184704) != 0):
                self.state = 72
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,4,self._ctx)
                if la_ == 1:
                    self.state = 70
                    self.attribute()
                    pass

                elif la_ == 2:
                    self.state = 71
                    self.method()
                    pass


                self.state = 76
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 77
            self.match(PlantUMLParser.T__4)
            self.state = 78
            self.match(PlantUMLParser.NL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AbstractContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return PlantUMLParser.RULE_abstract

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAbstract" ):
                listener.enterAbstract(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAbstract" ):
                listener.exitAbstract(self)




    def abstract(self):

        localctx = PlantUMLParser.AbstractContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_abstract)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 80
            self.match(PlantUMLParser.T__5)
            self.state = 82
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==3:
                self.state = 81
                self.match(PlantUMLParser.T__2)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class RelationshipContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def association(self):
            return self.getTypedRuleContext(PlantUMLParser.AssociationContext,0)


        def inheritance(self):
            return self.getTypedRuleContext(PlantUMLParser.InheritanceContext,0)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_relationship

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRelationship" ):
                listener.enterRelationship(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRelationship" ):
                listener.exitRelationship(self)




    def relationship(self):

        localctx = PlantUMLParser.RelationshipContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_relationship)
        try:
            self.state = 86
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,7,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 84
                self.association()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 85
                self.inheritance()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AssociationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.c_left = None # CardinalityContext
            self.c_right = None # CardinalityContext

        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(PlantUMLParser.ID)
            else:
                return self.getToken(PlantUMLParser.ID, i)

        def NL(self):
            return self.getToken(PlantUMLParser.NL, 0)

        def bidirectional(self):
            return self.getTypedRuleContext(PlantUMLParser.BidirectionalContext,0)


        def unidirectional(self):
            return self.getTypedRuleContext(PlantUMLParser.UnidirectionalContext,0)


        def aggregation(self):
            return self.getTypedRuleContext(PlantUMLParser.AggregationContext,0)


        def composition(self):
            return self.getTypedRuleContext(PlantUMLParser.CompositionContext,0)


        def cardinality(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PlantUMLParser.CardinalityContext)
            else:
                return self.getTypedRuleContext(PlantUMLParser.CardinalityContext,i)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_association

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAssociation" ):
                listener.enterAssociation(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAssociation" ):
                listener.exitAssociation(self)




    def association(self):

        localctx = PlantUMLParser.AssociationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_association)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 88
            self.match(PlantUMLParser.ID)
            self.state = 90
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==15:
                self.state = 89
                localctx.c_left = self.cardinality()


            self.state = 96
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,9,self._ctx)
            if la_ == 1:
                self.state = 92
                self.bidirectional()
                pass

            elif la_ == 2:
                self.state = 93
                self.unidirectional()
                pass

            elif la_ == 3:
                self.state = 94
                self.aggregation()
                pass

            elif la_ == 4:
                self.state = 95
                self.composition()
                pass


            self.state = 99
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==15:
                self.state = 98
                localctx.c_right = self.cardinality()


            self.state = 101
            self.match(PlantUMLParser.ID)
            self.state = 104
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==7:
                self.state = 102
                self.match(PlantUMLParser.T__6)
                self.state = 103
                self.match(PlantUMLParser.ID)


            self.state = 106
            self.match(PlantUMLParser.NL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class BidirectionalContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return PlantUMLParser.RULE_bidirectional

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBidirectional" ):
                listener.enterBidirectional(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBidirectional" ):
                listener.exitBidirectional(self)




    def bidirectional(self):

        localctx = PlantUMLParser.BidirectionalContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_bidirectional)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 108
            self.match(PlantUMLParser.T__7)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class UnidirectionalContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.nav_l = None # Token
            self.nav_r = None # Token


        def getRuleIndex(self):
            return PlantUMLParser.RULE_unidirectional

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterUnidirectional" ):
                listener.enterUnidirectional(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitUnidirectional" ):
                listener.exitUnidirectional(self)




    def unidirectional(self):

        localctx = PlantUMLParser.UnidirectionalContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_unidirectional)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 111
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==9:
                self.state = 110
                localctx.nav_l = self.match(PlantUMLParser.T__8)


            self.state = 113
            self.match(PlantUMLParser.T__7)
            self.state = 115
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==10:
                self.state = 114
                localctx.nav_r = self.match(PlantUMLParser.T__9)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AggregationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.aggr_l = None # Token
            self.aggr_r = None # Token


        def getRuleIndex(self):
            return PlantUMLParser.RULE_aggregation

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAggregation" ):
                listener.enterAggregation(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAggregation" ):
                listener.exitAggregation(self)




    def aggregation(self):

        localctx = PlantUMLParser.AggregationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_aggregation)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 123
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,16,self._ctx)
            if la_ == 1:
                self.state = 118
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==11:
                    self.state = 117
                    localctx.aggr_l = self.match(PlantUMLParser.T__10)


                pass

            elif la_ == 2:
                self.state = 121
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==9:
                    self.state = 120
                    self.match(PlantUMLParser.T__8)


                pass


            self.state = 125
            self.match(PlantUMLParser.T__7)
            self.state = 132
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,19,self._ctx)
            if la_ == 1:
                self.state = 127
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==10:
                    self.state = 126
                    self.match(PlantUMLParser.T__9)


                pass

            elif la_ == 2:
                self.state = 130
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==11:
                    self.state = 129
                    localctx.aggr_r = self.match(PlantUMLParser.T__10)


                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CompositionContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.comp_l = None # Token
            self.comp_r = None # Token

        def ASTK(self, i:int=None):
            if i is None:
                return self.getTokens(PlantUMLParser.ASTK)
            else:
                return self.getToken(PlantUMLParser.ASTK, i)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_composition

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComposition" ):
                listener.enterComposition(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComposition" ):
                listener.exitComposition(self)




    def composition(self):

        localctx = PlantUMLParser.CompositionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_composition)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 140
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,22,self._ctx)
            if la_ == 1:
                self.state = 135
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==37:
                    self.state = 134
                    localctx.comp_l = self.match(PlantUMLParser.ASTK)


                pass

            elif la_ == 2:
                self.state = 138
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==9:
                    self.state = 137
                    self.match(PlantUMLParser.T__8)


                pass


            self.state = 142
            self.match(PlantUMLParser.T__7)
            self.state = 149
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,25,self._ctx)
            if la_ == 1:
                self.state = 144
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==10:
                    self.state = 143
                    self.match(PlantUMLParser.T__9)


                pass

            elif la_ == 2:
                self.state = 147
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==37:
                    self.state = 146
                    localctx.comp_r = self.match(PlantUMLParser.ASTK)


                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class InheritanceContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.inh_left = None # Token

        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(PlantUMLParser.ID)
            else:
                return self.getToken(PlantUMLParser.ID, i)

        def NL(self):
            return self.getToken(PlantUMLParser.NL, 0)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_inheritance

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInheritance" ):
                listener.enterInheritance(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInheritance" ):
                listener.exitInheritance(self)




    def inheritance(self):

        localctx = PlantUMLParser.InheritanceContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_inheritance)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 151
            self.match(PlantUMLParser.ID)
            self.state = 154
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [12]:
                self.state = 152
                localctx.inh_left = self.match(PlantUMLParser.T__11)
                pass
            elif token in [13]:
                self.state = 153
                self.match(PlantUMLParser.T__12)
                pass
            else:
                raise NoViableAltException(self)

            self.state = 156
            self.match(PlantUMLParser.ID)
            self.state = 157
            self.match(PlantUMLParser.NL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ExtendsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_extends

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExtends" ):
                listener.enterExtends(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExtends" ):
                listener.exitExtends(self)




    def extends(self):

        localctx = PlantUMLParser.ExtendsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_extends)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 159
            self.match(PlantUMLParser.T__13)
            self.state = 160
            self.match(PlantUMLParser.ID)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CardinalityContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.min_ = None # CardinalityValContext
            self.max_ = None # CardinalityValContext

        def cardinalityVal(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PlantUMLParser.CardinalityValContext)
            else:
                return self.getTypedRuleContext(PlantUMLParser.CardinalityValContext,i)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_cardinality

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCardinality" ):
                listener.enterCardinality(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCardinality" ):
                listener.exitCardinality(self)




    def cardinality(self):

        localctx = PlantUMLParser.CardinalityContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_cardinality)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 162
            self.match(PlantUMLParser.T__14)
            self.state = 163
            localctx.min_ = self.cardinalityVal()
            self.state = 166
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==16:
                self.state = 164
                self.match(PlantUMLParser.T__15)
                self.state = 165
                localctx.max_ = self.cardinalityVal()


            self.state = 168
            self.match(PlantUMLParser.T__14)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CardinalityValContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT(self):
            return self.getToken(PlantUMLParser.INT, 0)

        def ASTK(self):
            return self.getToken(PlantUMLParser.ASTK, 0)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_cardinalityVal

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCardinalityVal" ):
                listener.enterCardinalityVal(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCardinalityVal" ):
                listener.exitCardinalityVal(self)




    def cardinalityVal(self):

        localctx = PlantUMLParser.CardinalityValContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_cardinalityVal)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 170
            _la = self._input.LA(1)
            if not(_la==36 or _la==37):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AttributeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def primitiveData(self):
            return self.getTypedRuleContext(PlantUMLParser.PrimitiveDataContext,0)


        def NL(self):
            return self.getToken(PlantUMLParser.NL, 0)

        def visibility(self):
            return self.getTypedRuleContext(PlantUMLParser.VisibilityContext,0)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_attribute

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAttribute" ):
                listener.enterAttribute(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAttribute" ):
                listener.exitAttribute(self)




    def attribute(self):

        localctx = PlantUMLParser.AttributeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_attribute)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 173
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 7864320) != 0):
                self.state = 172
                self.visibility()


            self.state = 175
            self.match(PlantUMLParser.ID)
            self.state = 176
            self.match(PlantUMLParser.T__6)
            self.state = 177
            self.primitiveData()
            self.state = 178
            self.match(PlantUMLParser.NL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class MethodContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def NL(self):
            return self.getToken(PlantUMLParser.NL, 0)

        def visibility(self):
            return self.getTypedRuleContext(PlantUMLParser.VisibilityContext,0)


        def modifier(self):
            return self.getTypedRuleContext(PlantUMLParser.ModifierContext,0)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_method

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMethod" ):
                listener.enterMethod(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMethod" ):
                listener.exitMethod(self)




    def method(self):

        localctx = PlantUMLParser.MethodContext(self, self._ctx, self.state)
        self.enterRule(localctx, 32, self.RULE_method)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 181
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 7864320) != 0):
                self.state = 180
                self.visibility()


            self.state = 184
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==31 or _la==32:
                self.state = 183
                self.modifier()


            self.state = 187
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==17:
                self.state = 186
                self.match(PlantUMLParser.T__16)


            self.state = 189
            self.match(PlantUMLParser.ID)
            self.state = 190
            self.match(PlantUMLParser.T__17)
            self.state = 191
            self.match(PlantUMLParser.NL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class VisibilityContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return PlantUMLParser.RULE_visibility

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVisibility" ):
                listener.enterVisibility(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVisibility" ):
                listener.exitVisibility(self)




    def visibility(self):

        localctx = PlantUMLParser.VisibilityContext(self, self._ctx, self.state)
        self.enterRule(localctx, 34, self.RULE_visibility)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 193
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 7864320) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PrimitiveDataContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return PlantUMLParser.RULE_primitiveData

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPrimitiveData" ):
                listener.enterPrimitiveData(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPrimitiveData" ):
                listener.exitPrimitiveData(self)




    def primitiveData(self):

        localctx = PlantUMLParser.PrimitiveDataContext(self, self._ctx, self.state)
        self.enterRule(localctx, 36, self.RULE_primitiveData)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 195
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 2139095040) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ModifierContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return PlantUMLParser.RULE_modifier

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterModifier" ):
                listener.enterModifier(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitModifier" ):
                listener.exitModifier(self)




    def modifier(self):

        localctx = PlantUMLParser.ModifierContext(self, self._ctx, self.state)
        self.enterRule(localctx, 38, self.RULE_modifier)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 197
            _la = self._input.LA(1)
            if not(_la==31 or _la==32):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








from antlr4 import CommonTokenStream, FileStream, ParseTreeWalker
from .PlantUMLLexer import PlantUMLLexer
from .PlantUMLParser import PlantUMLParser
from .plantUML_buml_listener import BUMLGenerationListener
from besser.BUML.metamodel.structural import DomainModel
import os

def plantuml_to_buml(plantUML_model_path:str, buml_model_file_name:str = "buml_model"):
    """Transforms a PlantUML model into a B-UML model.

    Args:
        plantUML_model_path (str): The path to the file containing the PlantUML code.
        buml_model_file_name (str, optional): the name of the file produced with the base code to build the B-UML model.

    Returns:
        BUML_model (DomainModel): the B-UML model object.
    """
    lexer = PlantUMLLexer(FileStream(plantUML_model_path))
    parser = PlantUMLParser(CommonTokenStream(lexer))
    parse_tree = parser.domainModel()
    # file creation
    if not os.path.exists("buml"):
        os.makedirs("buml")
    output = open("buml/" + buml_model_file_name + ".py","w+")
    listen = BUMLGenerationListener(output)
    walker = ParseTreeWalker()
    walker.walk(listen, parse_tree)
    output.close()
    # model creation
    namespace = {}
    with open("buml/" + buml_model_file_name + ".py", 'r') as model_code:
        code = model_code.read()
        exec(code, namespace)
    BUML_model: DomainModel = namespace.get('domain')
    return(BUML_model)


from .PlantUMLParser import PlantUMLParser
from .PlantUMLListener import PlantUMLListener

class BUMLGenerationListener(PlantUMLListener):
    visibility = {"+": "public",
                  "-": "private",
                  "#": "protected",
                  "~": "package"}

    def __init__(self, output):
        self.output = output
        self.__attr_list: list = list()
        self.__abstract_class: bool = False
        self.__dtypes: set = set()
        self.__classes: list = list()
        self.__relations: dict = dict()
        self.__ends: list = list()
        self.__inheritances: dict = dict()
        self.__relation_classes: list = list()
        self.__group_inh: int = 0
        self.__parent_classes: dict = dict()
        
    def enterClass(self, ctx: PlantUMLParser.ClassContext):
        text = "# " + ctx.ID().getText() + " class definition \n"
        self.output.write(text)
        self.__attr_list = []
        self.__abstract_class = False
    
    def exitClass(self, ctx: PlantUMLParser.ClassContext):
        attributes = list_to_str(self.__attr_list)
        text = ctx.ID().getText() + ": Class = Class(name=\"" + ctx.ID().getText() + "\", attributes=" + attributes
        if self.__abstract_class:
            text += ", is_abstract=True"
        text += ")\n\n"
        self.output.write(text)
        self.__classes.append(ctx.ID().getText())

    def enterAttribute(self, ctx: PlantUMLParser.AttributeContext):
        attribute_name = ctx.parentCtx.ID().getText() + "_" + ctx.ID().getText()
        text = attribute_name + ": Property = Property(name=\"" + ctx.ID().getText() + \
            "\", type="+ ctx.primitiveData().getText() +"_type"
        if ctx.visibility():
            text += ", visibility=\"" + self.visibility[ctx.visibility().getText()] + "\""
        text += ")\n"
        self.output.write(text)
        self.__attr_list.append(attribute_name)
        self.__dtypes.add(ctx.primitiveData().getText())

    def enterAbstract(self, ctx: PlantUMLParser.AbstractContext):
        self.__abstract_class = True
    
    def enterAssociation(self, ctx: PlantUMLParser.AssociationContext):
        self.__ends = []
            
    def exitAssociation(self, ctx: PlantUMLParser.AssociationContext):
        cl_name_1 = ctx.ID(0).getText()
        cl_name_2 = ctx.ID(1).getText()
        if ctx.ID(2) is None:
            raise ValueError("All the associations in the model must have a name")
        assoc_name = ctx.ID(2).getText()
        if assoc_name in self.__relations:
            raise ValueError("The model cannot have two associations with the same name")
        text = assoc_name + ": BinaryAssociation = BinaryAssociation(name=\"" + assoc_name + "\", ends={\n\
        Property(name=\"" + assoc_name + "\", type=" + cl_name_1 + ", multiplicity=" + getMultiplicity(ctx.c_left) + self.__ends[0] + "),\n\
        Property(name=\"" + assoc_name + "\", type=" + cl_name_2 + ", multiplicity=" + getMultiplicity(ctx.c_right) + self.__ends[1] + ")})\n"
        self.__relations[assoc_name] = text
        self.__relation_classes.append(cl_name_1)
        self.__relation_classes.append(cl_name_2)

    def enterBidirectional(self, ctx: PlantUMLParser.BidirectionalContext):
        self.__ends = ["", ""]
    
    def enterUnidirectional(self, ctx: PlantUMLParser.UnidirectionalContext):
        end_1 = ", is_navigable=True" if ctx.nav_l is not None else ", is_navigable=False"
        end_2 = ", is_navigable=True" if ctx.nav_r is not None else ", is_navigable=False"
        self.__ends.append(end_1)
        self.__ends.append(end_2)

    def enterComposition(self, ctx: PlantUMLParser.CompositionContext):
        end_1 = ", is_navigable=False, is_composite=True" if ctx.comp_l is not None else ""
        end_2 = ", is_navigable=False, is_composite=True" if ctx.comp_r is not None else ""
        self.__ends.append(end_1)
        self.__ends.append(end_2)

    def enterAggregation(self, ctx: PlantUMLParser.AggregationContext):
        end_1 = ", is_navigable=False, is_aggregation=True" if ctx.aggr_l is not None else ""
        end_2 = ", is_navigable=False, is_aggregation=True" if ctx.aggr_r is not None else ""
        self.__ends.append(end_1)
        self.__ends.append(end_2)
    
    def enterInheritance(self, ctx: PlantUMLParser.InheritanceContext):
        if ctx.inh_left:
            general = ctx.ID(0).getText()
            specific = ctx.ID(1).getText()
        else:
            general = ctx.ID(1).getText()
            specific = ctx.ID(0).getText()
        inheritance_name = "gen_" + general + "_" + specific
        text = inheritance_name + ": Generalization = Generalization(general=" + general + ", specific=" + specific + ")\n"
        self.__inheritances[inheritance_name] = text
        self.__relation_classes.append(general)
        self.__relation_classes.append(specific)

        if general not in self.__parent_classes:
            self.__parent_classes[general] = []
        self.__parent_classes[general].append(inheritance_name)

    def enterExtends(self, ctx: PlantUMLParser.ExtendsContext):
        general = ctx.ID().getText()
        specific = ctx.parentCtx.ID().getText()
        inheritance_name = "gen_" + general + "_" + specific
        text = inheritance_name + ": Generalization = Generalization(general=" + general + ", specific=" + specific + ")\n"
        self.__inheritances[inheritance_name] = text
    
    def enterSkinParam(self, ctx: PlantUMLParser.SkinParamContext):
        self.__group_inh = int(ctx.INT().getText())

    def exitDomainModel(self, ctx: PlantUMLParser.DomainModelContext):
        self.check_classes_definition()
        if len(self.__relations) != 0:
            self.output.write("# Relationships\n")
            for relation in self.__relations.values():
                self.output.write(relation)
        if len(self.__inheritances) != 0:
            self.output.write("\n# Generalizations\n")
            for inheritance in self.__inheritances.values():
                self.output.write(inheritance)
            if self.__group_inh > 1:
                self.create_generalization_set()
        classes = list_to_str(self.__classes)
        associations = list_to_str(list(self.__relations.keys()))
        generalizations = list_to_str(list(self.__inheritances.keys()))
        self.output.write("\n\n# Domain Model\n")
        self.output.write("domain: DomainModel = DomainModel(name=\"Domain Model\", types=" + classes + ", associations=" + associations + ", generalizations=" + generalizations + ")")
        text = '''from besser.BUML.metamodel.structural import NamedElement, DomainModel, Type, Class, \\
        Property, PrimitiveDataType, Multiplicity, Association, BinaryAssociation, Generalization, \\
        GeneralizationSet, AssociationClass \n\n'''
        text += "# Primitive Data Types \n"
        for dtype in self.__dtypes:
            text += dtype + "_type = PrimitiveDataType(\"" + dtype + "\")\n"
        text += "\n"
        self.output.seek(0)
        content = self.output.read()
        self.output.seek(0)
        self.output.write(text + content)

    def check_classes_definition(self):
        for cls in list(set(self.__relation_classes) - set(self.__classes)):
            text = "# " + cls + " class definition \n"
            text += cls + ": Class = Class(name=\"" + cls + "\", attributes={})\n\n"
            self.output.write(text)

    def create_generalization_set(self):
        for key, value in self.__parent_classes.items():
            if len(value) >= self.__group_inh:
                generalizations = ", ".join(value)
                text = key + "_generalization_set: GeneralizationSet = GeneralizationSet(name=\"" + key + \
                     "_gen_set\", generalizations={" + generalizations + "}, is_disjoint=True, is_complete=True)\n"
                self.output.write(text)
    
def list_to_str(list:list):
    if len(list) == 0:
        str_list = "set()"
    else:
        str_list = ", ".join(list)
        str_list = "{" + str_list + "}"
    return str_list

def getMultiplicity(cardinality:PlantUMLParser.CardinalityContext):
    min = ""
    max = ""
    multiplicity = ""
    if cardinality is None:
        min = "1"
        max = "1"
    else:
        if cardinality.cardinalityVal(0).INT():
            min = cardinality.cardinalityVal(0).INT().getText()
        elif cardinality.cardinalityVal(0).ASTK():
            min = "\"*\""
        if cardinality.cardinalityVal(1) and cardinality.cardinalityVal(1).INT():
            max = cardinality.cardinalityVal(1).INT().getText()
        elif cardinality.cardinalityVal(1) and cardinality.cardinalityVal(1).ASTK():
            max = "\"*\""
        if max == "":
            max = min
        if max == "\"*\"" == min:
            min = "0"

    multiplicity = "Multiplicity(" + min + ", " + max + ")"
    return (multiplicity)



from besser.BUML.notations.ocl.FactoryInstance import Factory
class Root_Handler:
    def __init__(self,ocl=None,dm=None,om=None):
        context = None
        if ocl is not None:
            context = ocl.context
        self.root = None
        self.dm = dm
        self.om = om
        self.factory = Factory(context)
        self.all =[]
        self.context = context
        self.invariant = False
        self.pre = False
        self.post = False
        self.if_else_roots = []

    def get_root(self):
        return self.root
    def get_inv(self):
        return self.invariant
    def set_inv(self,inv):
        self.invariant = inv
    def set_post(self,post):
        self.post = post
    def get_post(self):
        return self.post

    def set_body(self,body):
        self.body = body
    def get_body(self):
        return self.body
    def set_pre(self,pre):
        self.pre=pre
    def get_pre(self):
        return self.pre
    def create_property (self,prop):
        return self.factory.create_property_Call_Expression(prop,"NP")
    def handle_property(self,prop):
        self.add_to_root(self.factory.create_property_Call_Expression(prop,"NP"))

    def get_root(self):
        return self.root
    def get_context_name(self):
        return self.context_name
    def set_context_name(self,name):
        self.context_name = name
    def create_if_else_exp(self,name,type):
        self.if_else_roots.append(None)
        return self.factory.create_if_else_exp(name,type)
    def pop(self):
        if len(self.all)!=0:
            self.add_to_root(self.all.pop())
    def checkNumberOrVariable(self, txt):
        if txt.isnumeric():
            if "." in txt:
                return "real"
            else:
                return "int"
        elif "'" in txt:
            return "str"
        elif txt == "True" or txt == "False":
            return "bool"
        else:
            return "var"

    def _add_root(self,root,op):
        if len(self.all) == 0:
            if root is None:
                root = op
            else:
                op.source = root
                root = op
        else:
            self.all[-1].add_body(op)
        return root
    def getTop(self,last = False):
        currentHead = self.if_else_roots.pop()
        if not last:
            self.if_else_roots.append(None)
        return currentHead
    def add_to_root(self, op):
        self.last_added = op
        if len(self.if_else_roots) != 0:
            self.if_else_roots.append(self._add_root(self.if_else_roots.pop(),op))
            pass
        else:
            self.root = self._add_root(self.root, op)
        pass
    def pop_root(self,root):

        if self.root == self.last_coll_exp:
            self.root = None
            return self.last_coll_exp
        if hasattr(root, "arguments"):
            if self.last_coll_exp in root.arguments:
                root.arguments.remove(self.last_coll_exp)
                return self.last_coll_exp
            for args in root.arguments:
                self.pop_root(args)

    def print(self):
        self.handlePrint(self.root)
    def handle_ID(self,id):
        varID =self.factory.create_variable_expression(id,None)
        self.add_to_root(varID)
        # print('\x1b[6;30;42m' + 'handled ID, verify me!!!' + '\x1b[0m')
        pass
    def create_ordered_set(self):
        type = self.factory.create_ordered_set_type()
        return self.factory.create_collection_literal_expression("orderedSet", type)

    def create_set(self):
        type = self.factory.create_set_type()
        return self.factory.create_collection_literal_expression("set", type)

    def create_sub_ordered_set(self):
        type = self.factory.create_ordered_set_type()
        return self.factory.create_collection_literal_expression("subOrderedSet", type)
    def create_sequence(self):
        type = self.factory.create_sequence_type ()
        return self.factory.create_collection_literal_expression("sequence",type)
    def create_sub_sequence(self):
        type = self.factory.create_sub_sequence_type ()
        return self.factory.create_collection_literal_expression("subsequence",type)

    def create_bag(self):
        type = self.factory.create_bag_type()
        return self.factory.create_collection_literal_expression("bag",type = type)

    def create_type_exp(self,classifier):
        return self.factory.create_type_exp(classifier)

    def create_infinix_op(self,op):
        self.factory.create_infix_operator(op)
    def handle_and_with_function_call(self, text):
        op = None
        inF = None
        if text[0:3] == "and":
            inF = self.factory.create_infix_operator("AND")
            op = self.factory.create_operation_call_expression(name = "AND")
        if text[0:2] == "or":
            inF = self.factory.create_infix_operator("OR")
            op = self.factory.create_operation_call_expression(name = "OR")
        op.arguments.append(inF)
        self.add_to_root(op)
        pass
    def create_operation_call_exp(self,name):
        return self.factory.create_operation_call_expression(name=name)
    def get_factory(self):
        return self.factory
    def handle_bag(self,  collectionLiteral, operator):
        infixOperator = None
        if operator is not None:
            infixOperator = self.factory.create_infix_operator(operator)

        if infixOperator is not None:
            operationCallExp = self.factory.create_operation_call_expression(None,collectionLiteral,infixOperator,None,True)

        self.add_to_root(operationCallExp)
    def handle_adding_to_root(self, expression, op=None):
        if op is not None:
            expression.referredOperation(op)
        self.add_to_root(expression)
        pass
    def handlePrimaryExp(self,primaryExp,operator):
        pass

    def handle_collection(self,oclExp):

        collectionOperator = None
        if "forAll" in oclExp[0:8]:
            collectionOperator = "forAll"
            pass
        elif "exists" in oclExp[0:8]:
            collectionOperator = "exists"
            pass
        elif "collect" in oclExp[0:9]:
            collectionOperator = "collect"
            pass
        elif "select" in oclExp[0:8]:
            collectionOperator = "select"
            pass

        # print("Collection Operator: " + collectionOperator)
        self.handleColl(oclExp,collectionOperator)

        pass
    def handle_single_variable(self, variable_name,sign):
        op = self.factory.create_operation_call_expression(name = "operation")
        infinix_op = self.factory.create_infix_operator(sign)
        prop = self.factory.create_property_Call_Expression(variable_name,"NI")
        op.arguments.append(infinix_op)
        op.arguments.append(prop)
        self.add_to_root(op)
        pass
    def verify(self,item):
        referredOP= None
        if 'and' in item[0:3]:
            item = item[3:]
            referredOP= 'AND'
        if 'or' in item[0:2]:
            item = item[2:]
            referredOP = 'OR'
        prop = self.factory.create_property_Call_Expression(item, 'NI')
        if referredOP is None:
            # prop= self.factory.create_property_Call_Expression(item,'NI')
            self.add_to_root(prop)
        else:
            opCallExp = self.factory.create_operation_call_expression(name="and")
            opCallExp.referredOperation = self.factory.create_infix_operator(referredOP)
            self.add_to_root(opCallExp)
        pass
    def getClass(self,name):
        for type in self.dm.types:
            if name == type.name:
                return type
        raise Exception ("Class not found")
        pass
    def handleColl(self, forAllExp,collectionOperator):

        self.all.append(self.factory.create_loop_expression(collectionOperator))
        without_arrow = forAllExp.replace("->", '')
        without_collOp = without_arrow.replace(collectionOperator+"(", '')
        if "|" in without_collOp:
            iterator = without_collOp.split("|")[0]
            multiple_variable = iterator.split(',')
            for variable in multiple_variable:
                iteratorParts = variable.split(':')
                iteratorVariableName = iteratorParts[0]
                if ":" in variable:
                    iteratorclass =self.getClass(iteratorParts[1])
                else:
                    iteratorclass = "NotMentioned"
                iteratorExp = self.factory.create_iterator_expression(iteratorVariableName ,iteratorclass)
                self.all[-1].addIterator(iteratorExp)


        pass
    def handle_binary_expression(self, expression, operator,inbetween= None,beforeSign = None):
        expressionParts = expression.split(operator)

        leftside = self.checkNumberOrVariable(expressionParts[0])
        rightside = self.checkNumberOrVariable(expressionParts[1])

        leftPart = None
        rightPart = None

        if "var" in leftside:
            leftPart = self.factory.create_property_Call_Expression(expressionParts[0], type="NP",iterators=self.all)
        elif "int" in leftside:
            leftPart = self.factory.create_integer_literal_expression("NP", int(expressionParts[0]))
        elif "real" in leftside:
            leftPart = self.factory.create_real_literal_expression("NP", float(expressionParts[0]))
        elif "bool" in leftside:
            leftPart = self.factory.create_boolean_literal_expression("NP", bool(expressionParts[0]))
        elif "str" in leftside:
            leftPart = self.factory.create_string_literal_expression("str", expressionParts[0].replace("'",""))

        if "var" in rightside:
            rightPart = self.factory.create_property_Call_Expression(expressionParts[1], type="NP",iterators=self.all)
        elif "int" in rightside:
            rightPart = self.factory.create_integer_literal_expression("NP", int(expressionParts[1]))
        elif "real" in rightside:
            rightPart = self.factory.create_real_literal_expression("NP", float(expressionParts[1]))
        elif "bool" in rightside:
            rightPart = self.factory.create_boolean_literal_expression("NP", bool(expressionParts[1]))
        elif "str" in rightside:
            rightPart = self.factory.create_string_literal_expression("str", expressionParts[1].replace("'",""))

        infixOperator = self.factory.create_infix_operator(operator)
        beforeOp = None
        if beforeSign is not None:
            beforeOp = self.factory.create_infix_operator(beforeSign)
        inBetweenOp = None
        if inbetween is not None and len(inbetween)>0:
            inBetweenOp = self.factory.create_infix_operator(inbetween)
        opeartion_call_exp = self.factory.create_operation_call_expression(leftPart, rightPart, infixOperator,
                                                                           inBetweenOp,beforeOp)
        self.add_to_root(opeartion_call_exp)


        pass
    def handleBinaryFunc(self,operator,number):

        num = self.checkNumberOrVariable(number)

        if "int" in num:
            rightPart = self.factory.create_integer_literal_expression("NP", int(number))

        if "real" in num:
            rightPart = self.factory.create_real_literal_expression("NP", float(number))
        op = self.factory.create_infix_operator(operator)
        self.last_added.arguments.append(op)
        self.last_added.arguments.append(rightPart)

    def handle_last_opnum(self,operator,number):
        op= self.factory.create_operation_call_expression(name='callExp')
        op.arguments.append(self.factory.create_infix_operator(operator))
        num = self.checkNumberOrVariable(number)

        if "int" in num:
           rightPart = self.factory.create_integer_literal_expression("NP", int(number))

        if "real" in num:
            rightPart = self.factory.create_real_literal_expression("NP", float(number))

        op.arguments.append(
            rightPart
        )
        self.add_to_root(op)
    def handlePrint(self, root):
        if root == None:
            return
        if hasattr(root, 'arguments'):
            print(root.arguments)
            print(root.referredOperation())
            self.handlePrint(root.source())

        if hasattr(root,'body'):
            print(root.name)
            print(root.iterator)
            for item in root.body:
                print(item)




from antlr4 import CommonTokenStream, FileStream, ParseTreeWalker
from .deploymentLexer import deploymentLexer
from .deploymentParser import deploymentParser
from .depl_to_buml_listener import Deployment_BUML_Listener
from besser.BUML.metamodel.deployment import DeploymentModel
import os

def buml_deployment_model(deployment_textfile:str, buml_model_file_name:str = "deployment_buml_model"):
    """.

    Args:
        plantUML_model_path (str): The path to the file containing the deployment architecture textual definition.
        buml_model_file_name (str, optional): the name of the file produced with the base code to build the B-UML model.

    Returns:
        BUML_model (DeploymentModel): the B-UML model object.
    """
    lexer = deploymentLexer(FileStream(deployment_textfile))
    parser = deploymentParser(CommonTokenStream(lexer))
    parse_tree = parser.architecture()
    # file creation
    if not os.path.exists("buml"):
        os.makedirs("buml")
    output = open("buml/" + buml_model_file_name + ".py","w+")
    listen = Deployment_BUML_Listener(output)
    walker = ParseTreeWalker()
    walker.walk(listen, parse_tree)
    output.close()
    # model creation
    namespace = {}
    with open("buml/" + buml_model_file_name + ".py", 'r') as model_code:
        code = model_code.read()
        exec(code, namespace)
    BUML_model: DeploymentModel = namespace.get('deployment_model')
    return(BUML_model)


from besser.BUML.notations.ocl.BOCLLexer import BOCLLexer
from besser.BUML.notations.ocl.BOCLParser import BOCLParser
from besser.BUML.notations.ocl.BOCLListener import BOCLListener
from besser.BUML.notations.ocl.RootHandler import Root_Handler
from antlr4 import *

class OCLWrapper:

    def __init__(self,dm,om):
        self.dm = dm
        self.om = om

    # this method  parse the OCL expression and create the Tree
    # Arguments:
    # OCL Object that contains the expressions and the class
    def parse(self,ocl):

        try:
            input_stream = InputStream(ocl.expression)
            rootHandler = Root_Handler(ocl,self.dm,self.om)
            lexer = BOCLLexer(input_stream)
            stream = CommonTokenStream(lexer)
            parser = BOCLParser(stream)
            tree = parser.oclFile()
            listener = BOCLListener(rootHandler)
            walker = ParseTreeWalker()
            walker.walk(listener,tree)
            return True
        except Exception as e:
            print(e)
            return False
